/*
 * Copyright (c) 2003 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: boardlist_ui.c,v 1.36 2003/12/12 21:43:52 fuyu Exp $
 */

#include "config.h"

#include "ochusha_private.h"
#include "ochusha.h"
#include "ochusha_utils_2ch.h"
#include "worker.h"

#include "ochusha_ui.h"
#include "bbs_thread_ui.h"
#include "board_properties.h"
#include "boardlist_ui.h"
#include "boardlist_view.h"
#include "bulletin_board_ui.h"
#include "paned_notebook.h"
#include "response_editor.h"
#include "text_search_window.h"
#include "ugly_gtk2utils.h"

#include <glib.h>
#include <gtk/gtk.h>

#include <errno.h>

#if HAVE_ONIG_ONIGPOSIX_H
# include <onig/onigposix.h>
#else
# include <onigposix.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


#if DEBUG_WIDGET_MOST
static void select_category_cb(BoardlistView *view,
			       OchushaBoardCategory *category,
			       gpointer user_data);
#endif
static void category_expanded_cb(BoardlistView *view,
				 OchushaBoardCategory *category,
				 OchushaApplication *application);
static void category_collapsed_cb(BoardlistView *view,
				  OchushaBoardCategory *category,
				  OchushaApplication *application);
static void toggle_hide_category_cb(BoardlistView *view,
				    OchushaBoardCategory *category,
				    OchushaApplication *application);
static void toggle_hide_board_cb(BoardlistView *view,
				 OchushaBulletinBoard *board,
				 OchushaApplication *application);
static void kill_category_cb(BoardlistView *view,
			     OchushaBoardCategory *category,
			     OchushaApplication *application);
static void kill_board_cb(BoardlistView *view, OchushaBulletinBoard *board,
			  OchushaApplication *application);

static void boardlist_read_category_element_cb(OchushaBBSTable *table,
					       OchushaBoardCategory *category,
					       GHashTable *cat_attributes,
					       OchushaApplication *appl);
static void boardlist_read_board_element_cb(OchushaBBSTable *table,
					    OchushaBulletinBoard *board,
					    GHashTable *cat_attributes,
					    OchushaApplication *appl);

static void boardlist_write_category_element_cb(OchushaBBSTable *table,
						OchushaBoardCategory *category,
						FILE *boardlist_xml,
						OchushaApplication *appl);
static void boardlist_write_board_element_cb(OchushaBBSTable *table,
					     OchushaBulletinBoard *board,
					     FILE *boardlist_xml,
					     OchushaApplication *appl);

static void write_threadlist(gpointer data, gpointer user_data);

static void build_boardlist(WorkerThread *employee, gpointer args);

static void setup_boardlist(OchushaApplication *application);

static void filter_toggled_cb(GtkToggleButton *toggle_button,
			      OchushaApplication *application);
static void item_view_required_cb(PanedNotebook *paned_notebook,
				  gpointer item, GtkWidget **item_view,
				  const gchar **title, GtkWidget **tab_label,
				  OchushaApplication *application);
static void item_view_being_closed_cb(PanedNotebook *paned_notebook,
				      OchushaBulletinBoard *board,
				      GtkWidget *item_view,
				      OchushaApplication *application);
static void item_view_closed_cb(PanedNotebook *paned_notebook,
				OchushaBulletinBoard *board, GtkWidget *item_view,
				OchushaApplication *application);
static void page_switched_cb(PanedNotebook *paned_notebook,
			     GtkWidget *previous_page, GtkWidget *new_page,
			     OchushaApplication *application);

static void make_new_thread_button_cb(GtkWidget *widget,
				      OchushaApplication *application);
static void open_board_properties_button_cb(GtkWidget *widget,
					    OchushaApplication *application);
static void refresh_board_button_cb(GtkWidget *widget,
				    OchushaApplication *application);
static void start_threadlist_search_button_cb(GtkWidget *widget,
					      OchushaApplication *application);
static void new_thread_dialog_response_cb(GtkWidget *dialog, gint response_id,
					  OchushaApplication *application);
static GtkWidget *create_new_thread_dialog(OchushaApplication *application,
					   OchushaBulletinBoard *board);


/* Υå˰٤ǤⳫ(OchushaBulletinBoard *)Υꥹȡ*/
static GSList *board_list;

static pthread_mutex_t category_info_lock;
static pthread_mutex_t board_list_lock;


#define CATEGORY_INFO_LOCK				\
  if (pthread_mutex_lock(&category_info_lock) != 0)	\
    {							\
      fprintf(stderr, "Couldn't lock a mutex.\n");	\
      abort();						\
    }

#define CATEGORY_INFO_UNLOCK				\
  if (pthread_mutex_unlock(&category_info_lock) != 0)	\
    {							\
      fprintf(stderr, "Couldn't unlock a mutex.\n");	\
      abort();						\
    }


#define BOARD_LIST_LOCK					\
  if (pthread_mutex_lock(&board_list_lock) != 0)	\
    {							\
      fprintf(stderr, "Couldn't lock a mutex.\n");	\
      abort();						\
    }

#define BOARD_LIST_UNLOCK				\
  if (pthread_mutex_unlock(&board_list_lock) != 0)	\
    {							\
      fprintf(stderr, "Couldn't unlock a mutex.\n");	\
      abort();						\
    }


typedef struct _BoardCategoryGUIInfo
{
  gboolean expanded;
} BoardCategoryGUIInfo;


static GQuark category_gui_info_id;
static GQuark application_id;


void
prepare_boardlist_ui_initialization(OchushaApplication *application)
{
  if (pthread_mutex_init(&board_list_lock, NULL) != 0)
    {
      fprintf(stderr, "Couldn't init a mutex.\n");
      abort();
    }

  category_gui_info_id
    = g_quark_from_static_string("BoardlistUI::CategoryInfo");
  application_id
    = g_quark_from_static_string("BoardlistUI::Application");
}


static void
open_threadlist_view_cb(GtkWidget *selector, OchushaBulletinBoard *board,
			gboolean in_tab, gboolean with_browser,
			PanedNotebook *paned_notebook)
{
  if (!with_browser)
    paned_notebook_open_item_view(paned_notebook, board, in_tab);
  else
    {
      OchushaApplication *application
	= g_object_get_qdata(G_OBJECT(selector), application_id);
      if (application != NULL)
	ochusha_open_url(application,
			 ochusha_bulletin_board_get_base_url(board),
			 in_tab, with_browser);
#if 0	/* Թΰ */
      else
	fprintf(stderr, "open_threadlist_view_cb: %s\n",
		ochusha_bulletin_board_get_base_url(board));
#endif
    }
}


void
initialize_boardlist(OchushaApplication *application)
{
  GtkToolbar *toolbar;
  GtkWidget *scrolled_window;
  GtkWidget *filter_button;

  board_list = NULL;

  application->boardlist_view = boardlist_view_new();
  g_object_set_qdata(G_OBJECT(application->boardlist_view), application_id,
		     application);
  g_signal_connect(G_OBJECT(application->boardlist_view), "category_expanded",
		   G_CALLBACK(category_expanded_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view), "category_collapsed",
		   G_CALLBACK(category_collapsed_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view),
		   "toggle_hide_category",
		   G_CALLBACK(toggle_hide_category_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view),
		   "toggle_hide_board",
		   G_CALLBACK(toggle_hide_board_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view), "kill_category",
		   G_CALLBACK(kill_category_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view), "kill_board",
		   G_CALLBACK(kill_board_cb), application);
#if TRACE_MEMORY_USAGE
  g_signal_connect(G_OBJECT(application->boardlist_view), "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &application->boardlist_view);
#endif
  boardlist_view_set_default_open_in_tab(
				BOARDLIST_VIEW(application->boardlist_view),
				application->default_open_in_tab);
  gtk_widget_show(application->boardlist_view);

  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled_window),
				      GTK_SHADOW_ETCHED_IN);
  gtk_container_add(GTK_CONTAINER(scrolled_window),
		    application->boardlist_view);

  application->contents_window
    = paned_notebook_new_with_selector(application->boardlist_pane_style,
				       scrolled_window);
  g_signal_connect(G_OBJECT(application->boardlist_view),
		   "open_threadlist_view", G_CALLBACK(open_threadlist_view_cb),
		   application->contents_window);
  gtk_paned_set_position(GTK_PANED(application->contents_window),
			 application->boardlist_width);
  paned_notebook_set_always_show_contents_pane(PANED_NOTEBOOK(application->contents_window),
					       TRUE);
  paned_notebook_set_tab_shrinkable(PANED_NOTEBOOK(application->contents_window), application->board_tab_shrinkable);
  paned_notebook_set_minimum_tab_label_size(PANED_NOTEBOOK(application->contents_window), application->board_tab_minimum_size);
  paned_notebook_set_tooltips(PANED_NOTEBOOK(application->contents_window),
			      application->board_tab_enable_tooltips);

  toolbar = paned_notebook_get_toolbar(PANED_NOTEBOOK(application->contents_window));

  gtk_toolbar_prepend_space(toolbar);
  gtk_toolbar_insert_stock(toolbar, GTK_STOCK_NEW,
			   _("Make a new thread at the current board"),
			   "make_new_thread_at_current_board",
			   GTK_SIGNAL_FUNC(make_new_thread_button_cb),
			   application,
			   0);

  gtk_toolbar_prepend_space(toolbar);
  gtk_toolbar_insert_stock(toolbar, GTK_STOCK_PROPERTIES,
			   _("Open the current board's properties"),
			   "open_the_current_board_properties",
			   GTK_SIGNAL_FUNC(open_board_properties_button_cb),
			   application,
			   0);
  filter_button = gtk_toggle_button_new_with_label(_("Filter"));
  gtk_widget_show(filter_button);
  application->filter_button = GTK_TOGGLE_BUTTON(filter_button);
  gtk_toolbar_prepend_widget(toolbar, filter_button,
			     _("Apply filter to the current threadlist"),
			     "apply_filter_to_the_current_threadlist");
  g_signal_connect(G_OBJECT(filter_button), "toggled",
		   G_CALLBACK(filter_toggled_cb), application);
#if TRACE_MEMORY_USAGE
  g_signal_connect(G_OBJECT(filter_button), "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &application->filter_button);
#endif

  gtk_toolbar_prepend_space(toolbar);
  gtk_toolbar_insert_stock(toolbar, GTK_STOCK_REFRESH,
			   _("Refresh current threadlist"),
			   "refresh_current_threadlist",
			   GTK_SIGNAL_FUNC(refresh_board_button_cb),
			   application,
			   0);

  gtk_toolbar_insert_stock(toolbar, GTK_STOCK_FIND,
			   _("Start text search on current threadlist"),
			   "start_text_search_on_threadlist",
			   GTK_SIGNAL_FUNC(start_threadlist_search_button_cb),
			   application,
			   0);

#if DEBUG_WIDGET_MOST
  g_signal_connect(G_OBJECT(application->boardlist_view), "select_category",
		   G_CALLBACK(select_category_cb), application);
#endif
  g_signal_connect(G_OBJECT(application->contents_window),
		   "item_view_required", G_CALLBACK(item_view_required_cb),
		   application);
  g_signal_connect(G_OBJECT(application->contents_window),
		   "item_view_being_closed",
		   G_CALLBACK(item_view_being_closed_cb), application);
  g_signal_connect(G_OBJECT(application->contents_window),
		   "item_view_closed", G_CALLBACK(item_view_closed_cb),
		   application);
  g_signal_connect(G_OBJECT(application->contents_window),
		   "page_switched", G_CALLBACK(page_switched_cb),
		   application);

  application->threadlist_search_window = NULL;

  application->table = ochusha_bbs_table_new();

  /* γĥǥ¸ */
  g_signal_connect(G_OBJECT(application->table),
		   "boardlist_read_category_element",
		   G_CALLBACK(boardlist_read_category_element_cb),
		   application);
  g_signal_connect(G_OBJECT(application->table),
		   "boardlist_read_board_element",
		   G_CALLBACK(boardlist_read_board_element_cb),
		   application);
  g_signal_connect(G_OBJECT(application->table),
		   "boardlist_write_category_element",
		   G_CALLBACK(boardlist_write_category_element_cb),
		   application);
  g_signal_connect(G_OBJECT(application->table),
		   "boardlist_write_board_element",
		   G_CALLBACK(boardlist_write_board_element_cb),
		   application);

  if (ochusha_bbs_table_read_boardlist_xml(application->table,
					   &application->config))
    {
      setup_boardlist(application);
      return;
    }

  refresh_boardlist(application);
}


#if TRACE_MEMORY_USAGE
static void
trace_free(gpointer pointer)
{
  G_FREE(pointer);
}
#define TRACE_FREE	trace_free
#else
#define TRACE_FREE	g_free
#endif


static BoardCategoryGUIInfo *
ensure_board_category_info(OchushaBoardCategory *category)
{
  BoardCategoryGUIInfo *info;

  CATEGORY_INFO_LOCK
  {
    info = g_object_get_qdata(G_OBJECT(category), category_gui_info_id);
    if (info == NULL)
      {
	info = G_NEW0(BoardCategoryGUIInfo, 1);
	g_object_set_qdata_full(G_OBJECT(category), category_gui_info_id,
				info, (GDestroyNotify)TRACE_FREE);
      }
  }
  CATEGORY_INFO_UNLOCK;

  return info;
}


void
refresh_boardlist(OchushaApplication *application)
{
  WorkerJob *job = G_NEW0(WorkerJob, 1);

  job->canceled = FALSE;
  job->job = build_boardlist;
  job->args = application;

  commit_job(job);
}


void
finalize_boardlist(OchushaApplication *application)
{
  if (application->contents_window != NULL)
    {
      PanedNotebook *boardlist = PANED_NOTEBOOK(application->contents_window);
      GtkWidget *current_board;
      application->boardlist_width
	= gtk_paned_get_position(GTK_PANED(boardlist));
      current_board = paned_notebook_get_current_item_view(boardlist);

      if (current_board != NULL)
	application->threadlist_height
	  = gtk_paned_get_position(GTK_PANED(current_board));
    }

  if (application->table != NULL)
    {
      if (!ochusha_bbs_table_write_boardlist_xml(application->table,
						 &application->config))
	fprintf(stderr, "Couldn't write boardlist2.xml.\n");
    }

  g_slist_foreach(board_list, write_threadlist, application);
}


void
make_new_thread(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  OchushaBulletinBoard *board;
  BulletinBoardGUIInfo *info;
  GtkWidget *new_thread_dialog;

  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);

  if (application->config.offline)
    return;

  contents_window = PANED_NOTEBOOK(application->contents_window);

  board = (OchushaBulletinBoard *)paned_notebook_get_current_item(contents_window);
  if (board == NULL)
    return;

  if (!ochusha_bulletin_board_is_new_thread_supported(board))
    {
      ochusha_open_url(application,
		       ochusha_bulletin_board_get_base_url(board),
		       FALSE, TRUE);
      return;
    }

  info = ensure_bulletin_board_info(board, application);

  if (info->new_thread_dialog == NULL)
    {
      new_thread_dialog = create_new_thread_dialog(application, board);
      info->new_thread_dialog = new_thread_dialog;
    }
  else
    return;

  g_signal_connect(G_OBJECT(new_thread_dialog), "response",
		   G_CALLBACK(new_thread_dialog_response_cb), application);
  g_signal_connect(G_OBJECT(new_thread_dialog), "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &info->new_thread_dialog);

  gtk_widget_show_all(new_thread_dialog);
}


void
open_current_board_properties(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  OchushaBulletinBoard *board;

  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);

  contents_window = PANED_NOTEBOOK(application->contents_window);

  board = (OchushaBulletinBoard *)paned_notebook_get_current_item(contents_window);
  if (board == NULL)
    return;

  open_board_properties(application, board);
}


void
refresh_current_threadlist(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  OchushaBulletinBoard *board;

  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);

  contents_window = PANED_NOTEBOOK(application->contents_window);

  board = (OchushaBulletinBoard *)paned_notebook_get_current_item(contents_window);
  if (board == NULL)
    return;

  refresh_threadlist_view(application, board);
}


static void
refresh_threadlist_helper(gpointer data, gpointer user_data)
{
  OchushaBulletinBoard *board = (OchushaBulletinBoard *)data;
  OchushaApplication *application = (OchushaApplication *)user_data;
  refresh_threadlist_view(application, board);
  sleep(1);
}


void
refresh_current_threadlist_all(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);
  contents_window = PANED_NOTEBOOK(application->contents_window);
  paned_notebook_foreach_item(contents_window, refresh_threadlist_helper,
			      application);
}


static void
redraw_threadlist_helper(gpointer data, gpointer user_data)
{
  OchushaBulletinBoard *board = (OchushaBulletinBoard *)data;
  OchushaApplication *application = (OchushaApplication *)user_data;
  redraw_threadlist_view(application, board);
}


void
redraw_current_threadlist_all(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);
  contents_window = PANED_NOTEBOOK(application->contents_window);
  paned_notebook_foreach_item(contents_window, redraw_threadlist_helper,
			      application);
  fix_filter_button_state(application);
}


static ThreadlistSearchQueryData search_query_data =
{
  NULL,		/* key */
  TEXT_SEARCH_DIRECTION_DOWN,
  NULL,		/* last_view */
  {},		/* regexp */
  FALSE,	/* regexp_available */
  FALSE,	/* enable_wrap */
  FALSE,	/* match_case */
  FALSE,	/* use_regexp */
};


static void
text_search_window_response_cb(TextSearchWindow *window, gint response_id,
			       OchushaApplication *application)
{
  switch (response_id)
    {
    case GTK_RESPONSE_DELETE_EVENT:
      application->threadlist_search_window = NULL;
      /* fall through */

    case GTK_RESPONSE_CANCEL:
      /* XXX: ɥξȥϿ٤ */
      if (search_query_data.last_view != NULL)
	threadlist_view_set_being_searched(search_query_data.last_view, FALSE);
      search_query_data.last_view = NULL;
      gtk_widget_hide(GTK_WIDGET(window));
      return;
    }
}


static gboolean
do_search_on_current_threadlist(OchushaApplication *application,
				gboolean include_current_pos)
{
  GtkWidget *widget;
  ThreadlistView *view;

  g_return_val_if_fail(application != NULL
		       && application->contents_window != NULL, FALSE);

  widget = paned_notebook_get_current_item_view(
				PANED_NOTEBOOK(application->contents_window));
  if (widget == NULL)
    return FALSE;
  g_return_val_if_fail(IS_PANED_NOTEBOOK(widget), FALSE);

  widget = paned_notebook_get_selector(PANED_NOTEBOOK(widget));
  if (widget == NULL)
    return FALSE;

  g_return_val_if_fail(GTK_IS_SCROLLED_WINDOW(widget), FALSE);
  widget = gtk_bin_get_child(GTK_BIN(widget));

  g_return_val_if_fail(IS_THREADLIST_VIEW(widget), FALSE);
  view = THREADLIST_VIEW(widget);

  search_query_data.last_view = view;
  threadlist_view_set_being_searched(view, TRUE);

  return threadlist_view_find_next(view,
				   search_query_data.direction,
				   search_query_data.enable_wrap,
				   include_current_pos,
				   &search_query_data);
}


static gboolean
text_search_window_query_changed_cb(TextSearchWindow *window, const gchar *key,
				    TextSearchDirection direction,
				    gboolean enable_wrap, gboolean match_case,
				    gboolean use_regexp,
				    OchushaApplication *application)
{
  gboolean result;
  gchar *normalized_key;

  if (search_query_data.key != NULL)
    G_FREE(search_query_data.key);

  normalized_key = g_utf8_normalize(key, -1, G_NORMALIZE_ALL);
#if TRACE_MEMORY_USAGE
  {
    gchar *tmp_key = normalized_key;
    normalized_key = G_STRDUP(tmp_key);
    g_free(tmp_key);
  }
#endif
  if (!match_case)
    {
      gchar *case_normalized_key = g_utf8_casefold(normalized_key, -1);
      G_FREE(normalized_key);
#if TRACE_MEMORY_USAGE
      normalized_key = G_STRDUP(case_normalized_key);
      g_free(case_normalized_key);
#else
      normalized_key = case_normalized_key;
#endif
    }
  search_query_data.key = normalized_key;
  search_query_data.direction = direction;
  search_query_data.enable_wrap = enable_wrap;
  search_query_data.match_case = match_case;
  search_query_data.use_regexp = use_regexp;
  if (use_regexp)
    {
      if (search_query_data.regexp_available)
	regfree(&search_query_data.regexp);

      if (regcomp(&search_query_data.regexp,
		  normalized_key, REG_EXTENDED | REG_NEWLINE | REG_NOSUB) == 0)
	{
	  search_query_data.regexp_available = TRUE;
	  result = do_search_on_current_threadlist(application, TRUE);
	}
      else
	{
	  regfree(&search_query_data.regexp);
	  search_query_data.regexp_available = FALSE;
	  result = FALSE;
	}
    }
  else
    result = do_search_on_current_threadlist(application, TRUE);

  return result;
}


static gboolean
text_search_window_find_next_cb(TextSearchWindow *window, const gchar *key,
				TextSearchDirection direction,
				gboolean enable_wrap, gboolean match_case,
				gboolean use_regexp,
				OchushaApplication *application)
{
  return do_search_on_current_threadlist(application, FALSE);
}


static void
setup_threadlist_search_window(OchushaApplication *application)
{
  if (search_query_data.key != NULL)
    {
      G_FREE(search_query_data.key);
      search_query_data.key = NULL;
    }

  application->threadlist_search_window = text_search_window_new();
  text_search_window_set_enable_incremental_search(
		TEXT_SEARCH_WINDOW(application->threadlist_search_window),
		TRUE);
  g_signal_connect(application->threadlist_search_window, "response",
		   G_CALLBACK(text_search_window_response_cb),
		   application);
  g_signal_connect(application->threadlist_search_window, "query_changed",
		   G_CALLBACK(text_search_window_query_changed_cb),
		   application);
  g_signal_connect(application->threadlist_search_window, "find_next",
		   G_CALLBACK(text_search_window_find_next_cb),
		   application);
#if TRACE_MEMORY_USAGE
  g_signal_connect(application->threadlist_search_window, "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &application->threadlist_search_window);
#endif
  gtk_window_set_title(GTK_WINDOW(application->threadlist_search_window),
		       _("Find in the current threadlist"));
  gtk_window_set_transient_for(
			GTK_WINDOW(application->threadlist_search_window),
			application->top_level);
}


void
advance_threadlist_search(OchushaApplication *application,
			  ThreadlistView *view,
			  ThreadlistSearchAction search_action)
{
  switch (search_action)
    {
    case THREADLIST_SEARCH_ACTION_FORWARD:
    case THREADLIST_SEARCH_ACTION_REGEXP_FORWARD:
      search_query_data.direction = THREADLIST_SEARCH_DIRECTION_FORWARD;
      break;

    case THREADLIST_SEARCH_ACTION_BACKWARD:
    case THREADLIST_SEARCH_ACTION_REGEXP_BACKWARD:
      search_query_data.direction = THREADLIST_SEARCH_DIRECTION_BACKWARD;
      break;

    default:
      /* SHOULD NOT reached here */
      /* THREADLIST_SEARCH_ACTION_TERMINATEcaller¦ǽ */
      break;
    }

  if (application->threadlist_search_window == NULL
      || !GTK_WIDGET_VISIBLE(view)
      || search_query_data.last_view != view)
    {
      if (search_query_data.last_view != NULL)
	threadlist_view_set_being_searched(search_query_data.last_view,
					   FALSE);
      start_current_threadlist_search(application);
      search_query_data.last_view = view;
    }
  else
    do_search_on_current_threadlist(application,
				    search_query_data.last_view != view);
}


void
terminate_threadlist_search(OchushaApplication *application,
			    ThreadlistView *view)
{
  TextSearchWindow *window;
  window = TEXT_SEARCH_WINDOW(application->threadlist_search_window);
  if (window != NULL)
    g_object_ref(G_OBJECT(window));

  if (window != NULL)
    {
      text_search_window_response(window, GTK_RESPONSE_CANCEL);
      g_object_unref(G_OBJECT(window));
    }
}


void
start_current_threadlist_search(OchushaApplication *application)
{
  if (application->threadlist_search_window == NULL)
    setup_threadlist_search_window(application);

  text_search_window_set_key_selected(
		TEXT_SEARCH_WINDOW(application->threadlist_search_window));
  gtk_widget_show(application->threadlist_search_window);
  gtk_widget_grab_focus(application->threadlist_search_window);
}


/*
 * ϤʤΤѰդषOchushaBulletinBoardѲưŪ
 * ˻ΤƼΧŪб٤ʤΤʡ
 */
void
fix_filter_button_state(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  OchushaBulletinBoard *board;
  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);
  contents_window = PANED_NOTEBOOK(application->contents_window);
  board = (OchushaBulletinBoard *)paned_notebook_get_current_item(contents_window);
  if (board == NULL)
    return;

  gtk_toggle_button_set_active(application->filter_button,
		ensure_bulletin_board_info(board, application)->enable_filter);
}


void
update_threadlist_entry(OchushaApplication *application, OchushaBBSThread *thread)
{
  PanedNotebook *board_view;
  GtkWidget *scrolled_window;
  ThreadlistView *threadlist_view;
  PanedNotebook *boardlist = (PanedNotebook *)application->contents_window;

  g_return_if_fail(boardlist != NULL && thread != NULL);

  gdk_threads_enter();
  board_view = (PanedNotebook *)paned_notebook_get_item_view(boardlist,
							     thread->board);

  if (board_view == NULL)
    goto finish_update;	/* ĤɽƤʤ⤢ */

  scrolled_window = paned_notebook_get_selector(board_view);

  if (scrolled_window == NULL)
    {
      goto finish_update;	/* ե륿ߤǥɽƤʤ
				 * ⤢롣
				 */
    }

  threadlist_view
    = THREADLIST_VIEW(gtk_bin_get_child(GTK_BIN(scrolled_window)));

  update_threadlist_entry_style(application, threadlist_view, thread);

 finish_update:
  gdk_threads_leave();
}


static void
filter_toggled_cb(GtkToggleButton *toggle_button,
		  OchushaApplication *application)
{
  PanedNotebook *board_pane = PANED_NOTEBOOK(application->contents_window);
  OchushaBulletinBoard *board
    = (OchushaBulletinBoard *)paned_notebook_get_current_item(board_pane);
  BulletinBoardGUIInfo *info;

  if (board == NULL)
    return;

  info = ensure_bulletin_board_info(board, application);
  if (info->enable_filter == gtk_toggle_button_get_active(toggle_button))
    return;

  info->enable_filter = !info->enable_filter;

  gtk_toggle_button_set_active(toggle_button, info->enable_filter);
  gdk_threads_leave();
  redraw_threadlist_view(application, board);
  gdk_threads_enter();
}


#if DEBUG_WIDGET_MOST
static void
select_category_cb(BoardlistView *view, OchushaBoardCategory *category,
		   gpointer user_data)
{
  fprintf(stderr, "select_category_cb: %p, %p, %p\n",
	  view, category, user_data);
}
#endif


static void
item_view_required_cb(PanedNotebook *paned_notebook, gpointer item,
		      GtkWidget **item_view, const gchar **title,
		      GtkWidget **tab_label,
		      OchushaApplication *application)
{
  OchushaBulletinBoard *board;
  if (!OCHUSHA_IS_BULLETIN_BOARD(item))
    {
      *tab_label = NULL;
      *item_view = NULL;
      return;
    }

  board = OCHUSHA_BULLETIN_BOARD(item);
  *tab_label = get_tab_label(board->name);
  *item_view = open_bulletin_board(application, board, ICON_LABEL(*tab_label));

  if (*item_view == NULL)
    {
      gtk_object_sink(GTK_OBJECT(*tab_label));
      *tab_label = NULL;
      return;
    }

  BOARD_LIST_LOCK
  {
    if (g_slist_find(board_list, board) == NULL)
      board_list = g_slist_append(board_list, board);
  }
  BOARD_LIST_UNLOCK;

  gtk_widget_show(*item_view);
  *title = board->name;
}


static void
item_view_being_closed_cb(PanedNotebook *paned_notebook,
			  OchushaBulletinBoard *board,
			  GtkWidget *item_view,
			  OchushaApplication *application)
{
  OchushaAsyncBuffer *buffer
    = snatch_threadlist_source_buffer_for_board(board, application);

  /*
   * MEMO: ɽƤ륦åȤºݤ˼ΤƤ륿ߥ
   *       򤺤餹
   * 
   * GtkTreeViewŪidleؿϿƤꡢ줬Ƥ֤˥
   * åȤȲ줿åȤidleؿưƤޤ
   * ͡ˡʤΤǡǤϥåȤref_count1äƤ
   * Υץ饤ƥidleؿˤäơˡref_count򸺤餹
   * Υץ饤ƥidleؿưȤȤϡGTK+Ū
   * ȤäƤidleؿλŻλƤȤȤˤʤΤǡ
   * פ
   */
  g_object_ref(G_OBJECT(item_view));

  /*
   * MEMO: Ĥ褦ȤƤ륹Ρsubject.txtɤ߹ߡפ
   *       ֥󥰡פԤäƤthreadŪ˽λ롣
   */
  if (buffer != NULL)
    {
      gdk_threads_leave();

      ochusha_async_buffer_suspend(buffer);
      ochusha_async_buffer_terminate(buffer);
      ochusha_async_buffer_resume(buffer);

      gdk_threads_enter();
    }

  if (item_view != NULL)
    {
      if (paned_notebook->number_of_items_shown == 1)
	{
	  PanedNotebook *board_notebook = PANED_NOTEBOOK(item_view);
	  GtkWidget *threadlist_view
	    = gtk_bin_get_child(GTK_BIN(board_notebook->selector));
	  if (gtk_widget_is_focus(threadlist_view))
	    {
	      gtk_widget_grab_focus(GTK_WIDGET(application->boardlist_view));
	    }
	}
    }

  if (buffer != NULL)
    g_object_unref(G_OBJECT(buffer));
  return;
}


static void
item_view_closed_cb(PanedNotebook *paned_notebook, OchushaBulletinBoard *board,
		    GtkWidget *item_view, OchushaApplication *application)
{
  if (paned_notebook_get_current_item_view(paned_notebook) == NULL)
    {
      /* ɽˤǤƤڡĤ줿 */
      if (item_view != NULL && IS_PANED_NOTEBOOK(item_view))
	{
	  application->threadlist_height
	    = gtk_paned_get_position(GTK_PANED(item_view));
	}
    }

  finalize_bulletin_board_view(board, item_view);
  delayed_g_object_unref(G_OBJECT(item_view));
}


static void
page_switched_cb(PanedNotebook *paned_notebook,
		 GtkWidget *previous_page, GtkWidget *new_page,
		 OchushaApplication *application)
{
  g_return_if_fail((previous_page != NULL && IS_PANED_NOTEBOOK(previous_page))
		   || (new_page != NULL && IS_PANED_NOTEBOOK(new_page)));

  if (previous_page != NULL)
    {
      GtkWidget *scrolled_window;

      application->threadlist_height
	= gtk_paned_get_position(GTK_PANED(previous_page));
      scrolled_window
	= paned_notebook_get_selector(PANED_NOTEBOOK(previous_page));

      if (scrolled_window != NULL)
	{
	  ThreadlistView *threadlist_view
	    = THREADLIST_VIEW(gtk_bin_get_child(GTK_BIN(scrolled_window)));
	  if (search_query_data.last_view == threadlist_view)
	    {
	      threadlist_view_set_being_searched(threadlist_view, FALSE);
	      search_query_data.last_view = NULL;
	    }
	}
    }

  if (new_page != NULL)
    {
      OchushaBulletinBoard *board
	= (OchushaBulletinBoard *)paned_notebook_get_item(paned_notebook, new_page);
      BulletinBoardGUIInfo *info;
      GtkWidget *tmp_widget;

      g_return_if_fail(board != NULL);
      info = ensure_bulletin_board_info(board, application);
      paned_notebook_set_tooltips(PANED_NOTEBOOK(new_page),
				  application->thread_tab_enable_tooltips);
      paned_notebook_set_paned_style(PANED_NOTEBOOK(new_page),
				     application->threadlist_pane_style);
      gtk_paned_set_position(GTK_PANED(new_page),
			     application->threadlist_height);
      gtk_toggle_button_set_active(application->filter_button,
				   info->enable_filter);
      tmp_widget = paned_notebook_get_selector(PANED_NOTEBOOK(new_page));
      tmp_widget = gtk_bin_get_child(GTK_BIN(tmp_widget));
      threadlist_view_set_default_open_in_tab(THREADLIST_VIEW(tmp_widget),
					application->default_open_in_tab);
    }
}


static void
boardlist_read_category_element_cb(OchushaBBSTable *table,
				   OchushaBoardCategory *category,
				   GHashTable *category_attributes,
				   OchushaApplication *application)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);
  info->expanded = ochusha_utils_get_attribute_boolean(category_attributes,
						       "expanded");
  return;
}


static void
boardlist_read_board_element_cb(OchushaBBSTable *table,
				OchushaBulletinBoard *board,
				GHashTable *board_attributes,
				OchushaApplication *application)
{
  BulletinBoardGUIInfo *info = ensure_bulletin_board_info(board, application);
  info->show_mailto_mode
    = (ThreadViewMailtoMode)ochusha_utils_get_attribute_int(board_attributes,
							   "show_mailto_mode");
  if (info->show_mailto_mode < THREAD_VIEW_MAILTO_MODE_DEFAULT
      || info->show_mailto_mode > THREAD_VIEW_MAILTO_MODE_HIDE)
    info->show_mailto_mode = THREAD_VIEW_MAILTO_MODE_DEFAULT;

  info->enable_filter = ochusha_utils_get_attribute_boolean(board_attributes,
							    "enable_filter");
  info->filter.rule = ochusha_utils_get_attribute_string(board_attributes,
							 "filter.rule");
  info->filter.ignore_threshold
    = ochusha_utils_get_attribute_int(board_attributes,
				      "filter.ignore_threshold");
  info->last_name = ochusha_utils_get_attribute_string(board_attributes,
						       "last_name");
  info->last_mail = ochusha_utils_get_attribute_string(board_attributes,
						       "last_mail");

  info->board_local_a_bone
    = ochusha_utils_get_attribute_boolean(board_attributes,
					  "board_local_a_bone");
  info->a_bone_by_name
    = ochusha_utils_get_attribute_boolean(board_attributes, "a_bone_by_name");
  info->a_bone_by_name_pattern
    = ochusha_utils_get_attribute_string(board_attributes,
					 "a_bone_by_name_pattern");
  info->a_bone_by_id
    = ochusha_utils_get_attribute_boolean(board_attributes,
					  "a_bone_by_id");
  info->a_bone_by_id_pattern
    = ochusha_utils_get_attribute_string(board_attributes,
					 "a_bone_by_id_pattern");
  info->a_bone_by_content
    = ochusha_utils_get_attribute_boolean(board_attributes,
					  "a_bone_by_content");
  info->a_bone_by_content_pattern
    = ochusha_utils_get_attribute_string(board_attributes,
					 "a_bone_by_content_pattern");
  return;
}


#define OUTPUT_CATEGORY_ATTRIBUTE_BOOLEAN(file, category, attribute)	\
  fprintf(file,								\
	  "      <attribute name=\"" #attribute	"\">\n"			\
	  "        <boolean val=\"%s\"/>\n"				\
	  "      </attribute>\n",					\
	  (category)->attribute ? "true" : "false")


static void
boardlist_write_category_element_cb(OchushaBBSTable *table,
				    OchushaBoardCategory *category,
				    FILE *boardlist_xml,
				    OchushaApplication *application)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);

  OUTPUT_CATEGORY_ATTRIBUTE_BOOLEAN(boardlist_xml, info, expanded);
}


#define OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(file, board, attribute)		\
  fprintf(file,								\
	  "        <attribute name=\"" #attribute	"\">\n"		\
	  "          <boolean val=\"%s\"/>\n"				\
	  "        </attribute>\n",					\
	  (board)->attribute ? "true" : "false")

#define OUTPUT_BOARD_ATTRIBUTE_INT(file, board, attribute)		\
  fprintf(file,								\
	  "        <attribute name=\"" #attribute	"\">\n"		\
	  "          <int val=\"%d\"/>\n"				\
	  "        </attribute>\n", (board)->attribute)

#define OUTPUT_BOARD_ATTRIBUTE_STRING(file, board, attribute)		\
  do {									\
    if ((board)->attribute != NULL)					\
      {									\
	gchar *text = g_markup_escape_text((board)->attribute, -1);	\
	fprintf(file,							\
		"        <attribute name=\"" #attribute	"\">\n"		\
		"          <string>%s</string>\n"			\
		"        </attribute>\n", text);			\
	g_free(text);							\
      }									\
  } while (0)


static void
boardlist_write_board_element_cb(OchushaBBSTable *table,
				 OchushaBulletinBoard *board,
				 FILE *boardlist_xml,
				 OchushaApplication *application)
{
  BulletinBoardGUIInfo *info = ensure_bulletin_board_info(board, application);
  OUTPUT_BOARD_ATTRIBUTE_INT(boardlist_xml, info, show_mailto_mode);
  OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, enable_filter);
  OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info, last_name);
  OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info, last_mail);

  if (info->filter.rule != NULL)
    {
      OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info, filter.rule);
      OUTPUT_BOARD_ATTRIBUTE_INT(boardlist_xml, info, filter.ignore_threshold);
    }

  if (info->board_local_a_bone)
    {
      OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, board_local_a_bone);
      OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, a_bone_by_name);
      OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info,
				    a_bone_by_name_pattern);
      OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, a_bone_by_id);
      OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info, a_bone_by_id_pattern);
      OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, a_bone_by_content);
      OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info,
				    a_bone_by_content_pattern);
    }

  return;
}


static void
write_threadlist(gpointer data, gpointer user_data)
{
  OchushaBulletinBoard *board = (OchushaBulletinBoard *)data;
  OchushaApplication *application = (OchushaApplication *)user_data;

  /* XXX: 顼 */
  if (!ochusha_bulletin_board_write_threadlist_xml(board, &application->config))
    fprintf(stderr, "Couldn't write threadlist2.xml.\n");
}


static void
category_expanded_cb(BoardlistView *view, OchushaBoardCategory *category,
		     OchushaApplication *application)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);
  info->expanded = TRUE;
}


static void
category_collapsed_cb(BoardlistView *view, OchushaBoardCategory *category,
		      OchushaApplication *application)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);
  info->expanded = FALSE;
}


static void
toggle_hide_category_cb(BoardlistView *view, OchushaBoardCategory *category,
			OchushaApplication *application)
{
  category->hidden = !(category->hidden);
}


static void
toggle_hide_board_cb(BoardlistView *view, OchushaBulletinBoard *board,
		     OchushaApplication *application)
{
  board->hidden = !(board->hidden);
}


static void
kill_category_cb(BoardlistView *view, OchushaBoardCategory *category,
		 OchushaApplication *application)
{
  category->killed = TRUE;
}


static void
kill_board_cb(BoardlistView *view, OchushaBulletinBoard *board,
	      OchushaApplication *application)
{
  board->killed = TRUE;
}


static gboolean
prepare_category(OchushaBoardCategory *category, gpointer user_data)
{
  ensure_board_category_info(category);
  return TRUE;
}


static gboolean
prepare_board(OchushaBulletinBoard *board, gpointer user_data)
{
  ensure_bulletin_board_info(board, (OchushaApplication *)user_data);
  return TRUE;
}


static OchushaBulletinBoard *
process_board_move(OchushaBulletinBoard *old_board,
		   OchushaBulletinBoard *new_board,
		   gpointer user_data)
{
#if 0
  OchushaApplication *application = (OchushaApplication *)user_data;
#endif  
  /*
   * TODO: DATեʤɥå夷ƤǡΰžԤ٤
   *       İžɤߤGUIɲäɬפ⡣
   */
#if DEBUG_GUI
  fprintf(stderr, "process_board_move: old_board=%p new_board=%p\n",
	  old_board, new_board);
#endif
  return new_board;	/* ǤϿΤͥ */
}


void
redraw_boardlist(OchushaApplication *application)
{
  int status;

  if ((status = pthread_mutex_trylock(&board_list_lock)) != 0)
    {
      if (status == EBUSY)
	return;		/* İι1åɤ */

      fprintf(stderr, "Internal error: invalid mutex.\n");
      abort();
    }

  setup_boardlist(application);

  if (pthread_mutex_unlock(&board_list_lock) != 0)
    {
      fprintf(stderr, "Couldn't unlock a mutex.\n");
      abort();
    }
}


static void
build_boardlist(WorkerThread *employee, gpointer args)
{
  int status;
  OchushaAsyncBuffer *buffer;

  OchushaApplication *application = (OchushaApplication *)args;
  OchushaBBSTable *table = application->table;

  if ((status = pthread_mutex_trylock(&board_list_lock)) != 0)
    {
      if (status == EBUSY)
	return;		/* İι1åɤ */

      fprintf(stderr, "Internal error: invalid mutex.\n");
      abort();
    }

  buffer = ochusha_utils_2ch_get_bbsmenu(application->broker);

  if (buffer == NULL)
    {
      /* XXX: 顼 */
      fprintf(stderr, "Couldn't get %s\n", application->config.bbsmenu_url);
      goto cleanup_before_return;
      return;
    }

  if (ochusha_utils_2ch_analyze_bbsmenu(table, buffer,
					prepare_category,
					prepare_board,
					process_board_move,
					application))
    {
      gdk_threads_enter();
      setup_boardlist(application);
      gdk_threads_leave();

      ochusha_bbs_table_write_boardlist_xml(table, &application->config);
    }

 cleanup_before_return:
  if (buffer != NULL)
    g_object_unref(G_OBJECT(buffer));

  if (pthread_mutex_unlock(&board_list_lock) != 0)
    {
      fprintf(stderr, "Couldn't unlock a mutex.\n");
      abort();
    }
}


static void
restore_expanded_state(OchushaBoardCategory *category, BoardlistView *view)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);
  if (info->expanded)
    boardlist_view_expand_category(view, category);
}


static void
setup_boardlist(OchushaApplication *application)
{
  boardlist_view_open(BOARDLIST_VIEW(application->boardlist_view),
		      application->table->category_list,
		      application->hide_hidden_boardlist_items);
  g_slist_foreach(application->table->category_list,
		  (GFunc)restore_expanded_state, application->boardlist_view);

  gtk_widget_show_all(application->boardlist_view);
  gtk_widget_show_all(application->contents_window);
}


static void
make_new_thread_button_cb(GtkWidget *widget, OchushaApplication *application)
{
  make_new_thread(application);
}


static void
open_board_properties_button_cb(GtkWidget *widget,
				OchushaApplication *application)
{
  open_current_board_properties(application);
}


static void
refresh_board_button_cb(GtkWidget *widget, OchushaApplication *application)
{
  refresh_current_threadlist(application);
}


static void
start_threadlist_search_button_cb(GtkWidget *widget,
				  OchushaApplication *application)
{
  start_current_threadlist_search(application);
}


static void
new_thread_dialog_response_cb(GtkWidget *dialog, gint response_id,
			      OchushaApplication *application)
{
  OchushaBulletinBoard *board = g_object_get_data(G_OBJECT(dialog), "board");
  gboolean result;

  if (response_id == GTK_RESPONSE_OK)
    {
      ResponseEditor *editor = g_object_get_data(G_OBJECT(dialog), "editor");
      GtkEntry *title_entry = g_object_get_data(G_OBJECT(dialog), "title");
      const gchar *title = gtk_entry_get_text(title_entry);
      gchar *message;
      OchushaBBSResponse response;

      response.name = response_editor_get_name(editor);
      response.mailto = response_editor_get_mail(editor);
      message = response_editor_get_response(editor);
      response.content = message;

      if (title != NULL && *title != '\0'
	  && message != NULL && *message != '\0')
	{
	  result = ochusha_bulletin_board_create_new_thread(board,
						application->broker,
						title, &response);
	  g_free(message);
	  if (result)
	    {
	      BulletinBoardGUIInfo *board_info
		= ensure_bulletin_board_info(board, application);

	      if (board_info->last_name == NULL
		  || strcmp(board_info->last_name, response.name) != 0)
		{
		  if (board_info->last_name != NULL)
		    G_FREE(board_info->last_name);
		  board_info->last_name = G_STRDUP(response.name);
		}

	      if (board_info->last_mail == NULL
		  || strcmp(board_info->last_mail, response.mailto) != 0)
		{
		  if (board_info->last_mail != NULL)
		    G_FREE(board_info->last_mail);
		  board_info->last_mail = G_STRDUP(response.mailto);
		}

	      if (application->last_name == NULL
		  || strcmp(application->last_name, response.name) != 0)
		{
		  if (application->last_name != NULL)
		    G_FREE(application->last_name);
		  application->last_name = G_STRDUP(response.name);
		}

	      if (application->last_mail == NULL
		  || strcmp(application->last_mail, response.mailto) != 0)
		{
		  if (application->last_mail != NULL)
		    G_FREE(application->last_mail);
		  application->last_mail = G_STRDUP(response.mailto);
		}
	    }
	}
      else
	{
	  /* ʸʤϥ顼ˤʤ餷 */
	  result = TRUE;
	  response_id = GTK_RESPONSE_CANCEL;
	}
    }
  else
    result = TRUE;

  if (result)
    {
      if (response_id == GTK_RESPONSE_OK)
	refresh_threadlist_view(application, board);
      gtk_widget_destroy(dialog);
    }
  else
    ochusha_open_url(application,
		     ochusha_bulletin_board_get_base_url(board),
		     FALSE, TRUE);
}


static GtkWidget *
create_new_thread_dialog(OchushaApplication *application,
			 OchushaBulletinBoard *board)
{
  GtkWidget *new_thread_dialog;
  GtkWidget *dialog_vbox;
#if 0
  GtkWidget *url_hbox;
  GtkWidget *url_entry;
#endif
  GtkWidget *title_hbox;
  GtkWidget *title_entry;
  GtkWidget *board_label;
  GtkWidget *editor;
  ResponseEditor *response_editor;
  BulletinBoardGUIInfo *board_info = ensure_bulletin_board_info(board,
								application);
  gboolean has_name = FALSE;
  gboolean has_mail = FALSE;
  gchar board_name_buffer[2048];

  new_thread_dialog = gtk_dialog_new_with_buttons(_("Create a New Thread"),
					 application->top_level,
					 GTK_DIALOG_DESTROY_WITH_PARENT,
					 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					 GTK_STOCK_OK, GTK_RESPONSE_OK,
					 NULL);
  gtk_window_set_position(GTK_WINDOW(new_thread_dialog), GTK_WIN_POS_CENTER);
  gtk_window_set_default_size(GTK_WINDOW(new_thread_dialog), -1, 250);

  dialog_vbox = GTK_DIALOG(new_thread_dialog)->vbox;

  snprintf(board_name_buffer, 2048, _("Create a New Thread on %s board"),
	   ochusha_bulletin_board_get_name(board));
  board_label = gtk_label_new(board_name_buffer);
  gtk_label_set_justify(GTK_LABEL(board_label), GTK_JUSTIFY_LEFT);
  gtk_widget_show(board_label);
  gtk_box_pack_start(GTK_BOX(dialog_vbox), board_label, FALSE, FALSE, 10);

#if 0
  url_hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(url_hbox), gtk_label_new("URL: "),
		     FALSE, FALSE, 0);
  url_entry = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(url_entry),
		     ochusha_bulletin_board_get_base_url(board));
  gtk_editable_set_editable(GTK_EDITABLE(url_entry), FALSE);
  GTK_WIDGET_UNSET_FLAGS(url_entry, GTK_CAN_FOCUS);
  gtk_box_pack_start(GTK_BOX(url_hbox), url_entry, TRUE, TRUE, 0);
  gtk_widget_show_all(url_hbox);
  gtk_box_pack_start(GTK_BOX(dialog_vbox), url_hbox, FALSE, FALSE, 0);
#endif

  title_hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(title_hbox), gtk_label_new(_("Title: ")),
		     FALSE, FALSE, 0);
  title_entry = gtk_entry_new();
  gtk_editable_set_editable(GTK_EDITABLE(title_entry), TRUE);
  GTK_WIDGET_SET_FLAGS(title_entry, GTK_CAN_FOCUS);
  gtk_box_pack_start(GTK_BOX(title_hbox), title_entry, TRUE, TRUE, 0);
  gtk_widget_show_all(title_hbox);
  gtk_box_pack_start(GTK_BOX(dialog_vbox), title_hbox, FALSE, FALSE, 0);

  editor = response_editor_new();
  response_editor = RESPONSE_EDITOR(editor);

  if (board_info->last_name != NULL)
    {
      response_editor_add_name(response_editor, board_info->last_name);
      if (!has_name)
	{
	  response_editor_set_name(response_editor, board_info->last_name);
	  has_name = TRUE;
	}
    }

  if (board_info->last_mail != NULL)
    {
      response_editor_add_mail(response_editor, board_info->last_mail);
      if (!has_mail)
	{
	  response_editor_set_mail(response_editor, board_info->last_mail);
	  has_mail = TRUE;
	}
    }

  if (application->last_name != NULL)
    {
      response_editor_add_name(response_editor, application->last_name);
      if (!has_name)
	response_editor_set_name(response_editor, application->last_name);
    }

  if (application->last_mail != NULL)
    {
      response_editor_add_mail(response_editor, application->last_mail);
      if (!has_mail)
	response_editor_set_mail(response_editor, application->last_mail);
    }

  gtk_widget_show(editor);

  gtk_box_pack_start(GTK_BOX(dialog_vbox), editor, TRUE, TRUE, 0);

  g_object_set_data(G_OBJECT(new_thread_dialog), "editor", editor);
  g_object_set_data(G_OBJECT(new_thread_dialog), "title", title_entry);
  g_object_set_data(G_OBJECT(new_thread_dialog), "board", board);

  return new_thread_dialog;
}

